--
local versionNumber = 1
local fileModified = true -- set this to true if you change this file for your scenario
-- if another file requires this file, it checks the version number to ensure that the
-- version is recent enough to have all the expected functionality
-- if you set fileModified to true, the error generated if this file is out of date will
-- warn you that you've modified this file
--
--
-- Settings for the navy module
local object = require("object")
local gen = require("generalLibrary"):minVersion(1)
local navy = require("navy"):minVersion(1)
local readRules = require("readRules"):minVersion(1)


-- This module implements features related to ships:
--  When can ships unload onto beaches
--  What planes can use which carriers
--

-- shipCapacityList[unitType.id] = nil or integer
-- supplies the default carrying capacity of each ship
-- if nil, the unit always has a capacity of 0

local shipCapacityList = navy.buildShipCapacityListFrom(readRules.loadedRules)
navy.registerShipCapacityList(shipCapacityList)



-- Unloading Units onto beaches
-- Both the ship doing the unloading and the unit
-- being unloaded must satisfy these conditions
-- in order to be activated
-- Change the @LANDFALL message so there is no option
-- to unload a ship by sailing into a land tile
--[[
@LANDFALL
@title=Disembark
Shall we disembark, Sire,
and leave the ships behind?

Stay With Ships
Make Landfall
--
--]]
-- Loading Units onto ships:
-- The unit and the ship type must both satisfy the
-- conditions, which are determined based on the location
-- of the unit attempting to board the ship
-- (if the unit can't board a ship type, the type's carrying capacity
-- will be set to 0)
--These restrictions may not hold for the AI
-- beachShipSettings[unitType.id] = {
--  .restrictBeachUnload = boolean or number or nil
--      if false or nil, unit can unload onto any beach
--      if true, unit can never unload onto a "beach"
--      (but might be able to unload beside a foreign port, see below)
--      if 0 or positive number, the unit can not have expended
--      more than that many "full" movement points.  That is, can only unload if
--      unit.moveSpent <= beachShipSettings[unit.type.id].restricBeachUnload*totpp.movementMultipliers.aggregate
--      if negative number, the unit can not have expended more than
--      that many 'atomic' movement points, that is
--      unit.moveSpent <= -beachShipSettings[unit.type.id].restricBeachUnload
--
--  .restrictPeacePortUnload = boolean or number or nil
--      if false or nil, the unit can unload from a square adjacent to a city
--      owned by a civ with which the player is at peace (or allied) without penalty,
--      even if restrictBeachUnload is true
--      if true, the unit can only unload if it can unload onto a "beach"
--      if 0 or positive number, the unit can not have expended
--      more than that many "full" movement points.  That is, can only unload if
--      unit.moveSpent <= beachShipSettings[unit.type.id].restricPeacePortUnload*totpp.movementMultipliers.aggregate
--      if negative number, the unit can not have expended more than
--      that many 'atomic' movement points, that is
--      unit.moveSpent <= -beachShipSettings[unit.type.id].restricPeacePortUnload
--
--  .restrictAlliedPortUnload = boolean or number or nil
--      if false or nil, the unit can unload from a square adjacent to a city
--      owned by a civ with which the player is allied without penalty,
--      even if restrictBeachUnload and restrictPortUnload are true
--      if true, the unit can only unload if it can unload onto a "beach" or peace port
--      if 0 or positive number, the unit can not have expended
--      more than that many "full" movement points.  That is, can only unload if
--      unit.moveSpent <= beachShipSettings[unit.type.id].restricAlliedPortUnload*totpp.movementMultipliers.aggregate
--      if negative number, the unit can not have expended more than
--      that many 'atomic' movement points, that is
--      unit.moveSpent <= -beachShipSettings[unit.type.id].restricAlliedPortUnload
--
--  .beachUnloadPenalty = function(unit)-->void or number or nil
--      if nil, nothing is done when unloading the unit
--      if number is 0 or positive, set the moveSpent of that unit so that
--      that many full movement points are expended, unless more have already been expended
--      i.e.
--      unit.moveSpent = math.max(unit.moveSpent, beachShipSettings[unit.type.id].beachUnloadPenalty*totpp.movementMultipliers.aggregate)
--      if number is negative, set the move spent so that many atomic movement units have been spent
--      if function, that function will be run for the unit unloaded, then for the ship carrying it
--      If different behaviour is desired for unloading beside a port, check for that and change the
--      penalty applied within the function
--
--  .unloadCheckFunction = nil or function(unit) --> boolean
--      if function, the unit can only unload if the function returns true
--      (and all other conditions are met)
--      nil means no extra condition
--  
--  .unloadFailureMessage = nil or string or function(unit) --> string
--      Message to show if the unit can't unload.
--      The message for the ship is shown at the top of the message
--      box, and the message for the cargo is shown below that
--      if nil, a generic message is shown.
--      If string, the string is shown, with %STRING1 replaced by the unit
--      type name.
--      If function, the returned string is used for the message, with
--      %STRING1 replaced by the unit type name
--
--  .forbidBeachBoarding = nil or boolean
--      If true, a unit can't get onto a ship from the "beach",
--      unless that unit is adjacent to a city at peace or allied
--      with the unit's owner
--      false or nil means it can

--  .forbidPeacePortBoarding = nil or boolean
--      If true, a unit can't get onto a ship from a tile
--      adjacent to a city with which it is at peace with
--      (unless forbidBeachBoarding isn't active)
--      false or nil means the unit can load from a tile
--      adjacent to a city with which the owner is at peace
--
--  .forbidAlliedPortBoarding = nil or boolean
--      if true, a unit can't get onto a ship from a tile
--      adjacent to an ally's city (unless it is allowed to
--      board a ship from a beach or peace port)
--
--  .boardingCheckFunction = nil or function(boardingUnit,shipType) --> boolean
--      if nil, ignore this condition
--      if a function and it returns true, the unit can board (if other conditions
--      allow it).  If the function returns false, the unit can't board
--      The boardingUnit is the unit checking if it can board a ship, and
--      the ship type is the unit type that may be boarded
--
--  forbidBoarding = nil or table of unit types
--      units in this table (if it is supplied) can't board this transport
--      type, or this type of unit can't board the transport types listed
--      (depending on whether this setting is for a land unit or 
--      sea transport)
--      This is achieved by clearing the sleep status of any invalid cargo
--      when a ship is activated within a city (as well as forbidding the
--      other types of boarding)
--

local beachShipSettings = {}

-- Here are some testing examples
beachShipSettings[object.uWacoGlider.id] = {
    restrictBeachUnload = true,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = true,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
	-- Waco Glider can only carry British 1st Airborne, US 75mm Pack How., US 82nd Airborne and US 75mm Pack How. units
	forbidBoarding = {object.uEngineer,
		object.uGIVeteran,
		object.uGIRegular,
		object.uUSBazooka,
		object.uRanger, 
		object.uDUKW,     
		object.uArmoredInfantry,
		object.uM5Stuart,       
		object.uM4A1Sherman,    
		object.uM4A376WSherman, 
		object.uM10Wolverine,   
		object.uM7105mmHMC,     
		object.uM12155mmHMC,    
		object.uUSBofors40mmAA, 
		object.uM2A1105mmArt,   
		object.uM1155mmArt,     
		object.uUSTruck,        
		object.uBritishVeteran,
		object.uBritishRegular,
		object.uBritishPIAT,   
		object.uBritishMG,     
		object.uCommandos, 
		object.uDaimlerAC,     
		object.uBritishMotorTroops, 
		object.uShermanIIaBde,		
		object.uShermanIIa,    
		object.uChurchillMkIV, 
		object.uShermanVCFirefly,
		object.uGurkhas,       
		object.uBishop25pdr,   
		object.uUKBofors40mmAA,
		object.u25pdrMKIIArt,  
		object.u55inchArt,     
		object.uUKTruck, 
		object.uCdnMotInf,
		object.uCdnShermanIII,
		object.uCdnM7105mm,
		object.uCdnMotorTroops,
		object.uCdnInfantry,
		object.uRadarStation,
		object.uSAMotInf,
		object.uSAShermanIII,
		object.uSASexton,
		object.uSAMotorTroops,
		object.uBraz1stInf,
		object.uGurkhas,
		object.uIndianInfantry,
		object.u4thIndianDiv,
		object.uIndian25pdrArt,
		object.uPolishInfantry,
		object.uPolishShermanIII,
		object.uPolish25pdrArt,
		object.uInfanterieColoniale,
		object.uGoumierMarocain,
		object.uTirailleurAlgerien,
		object.uFreeFrenchInfantry,
		object.uFreeFrench105Art,
		object.uFreeFrenchM10,
		object.uNZMotInf,
		object.uNZShermanIII,
		object.uNZ25pdrArt,
		object.uChurchillAVRE,
		object.uUS10thMountain,
		object.uUS1stInfantry,
		object.uCromwellMkVII,
		object.uM4A1ShermanInd,
		object.uGreekMtnBde,
		object.uJewishBde,
		object.uDocks,
		object.uM157mmAT,
		object.u6pdrAT,
		object.uDocks,			
		object.uSupplyHub,
		object.uBritish52ndMtn,
	},
    unloadFailureMessage = "%STRING1 units can only disembark units within a city or from a beachhead tile."
}

beachShipSettings[object.uC47Dakota.id] = {
    restrictBeachUnload = true,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = true,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
	-- C47 Dakota can only carry infantry type units
	forbidBoarding = {object.uEngineer,
		object.uDUKW,     
		object.uArmoredInfantry,
		object.uM5Stuart,       
		object.uM4A1Sherman,    
		object.uM4A376WSherman, 
		object.uM10Wolverine,   
		object.uM7105mmHMC,     
		object.uM12155mmHMC,    
		object.uUSBofors40mmAA, 
		object.uM2A1105mmArt,   
		object.uM1155mmArt,     
		object.uUSTruck,        
		object.uDaimlerAC,     
		object.uBritishMotorTroops, 
		object.uShermanIIaBde,		
		object.uShermanIIa,    
		object.uChurchillMkIV, 
		object.uShermanVCFirefly,
		object.uBishop25pdr,   
		object.uUKBofors40mmAA,
		object.u25pdrMKIIArt,  
		object.u55inchArt,     
		object.uUKTruck,
		object.uCdnMotInf,
		object.uCdnShermanIII,
		object.uCdnM7105mm,
		object.uCdnMotorTroops,
		object.uRadarStation,
		object.uSAMotInf,
		object.uSAShermanIII,
		object.uSASexton,
		object.uSAMotorTroops,
		object.uIndian25pdrArt,
		object.uPolishShermanIII,
		object.uPolish25pdrArt,
		-- object.uInfanterieColoniale,
		-- object.uGoumierMarocain,
		-- object.uTirailleurAlgerien,
		object.uFreeFrenchInfantry,
		object.uFreeFrench105Art,
		object.uFreeFrenchM10,
		object.uNZMotInf,
		object.uNZShermanIII,
		object.uNZ25pdrArt,
		object.uChurchillAVRE,
		object.uCromwellMkVII,
		object.uM4A1ShermanInd,
		object.uDocks,
		object.uSupplyHub,
	},
    unloadFailureMessage = "%STRING1 units can only disembark units within a city or from a beachhead tile."
}

beachShipSettings[object.uYorkClassCA.id] = {
    restrictBeachUnload = false,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = false,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
	-- York Class CA can only carry British 1st Airborne and Allied Generals units
	forbidBoarding = {object.uEngineer,
		object.uGIVeteran,
		object.uGIRegular,
		object.uUSBazooka,
		object.uRanger, 
		object.uDUKW,     
		object.uArmoredInfantry,
		object.uM5Stuart,       
		object.uM4A1Sherman,    
		object.uM4A376WSherman, 
		object.uM10Wolverine, 
		object.uRanger,
		object.u82ndAirborne,
		object.uM157mmAT,		
		object.uM7105mmHMC,     
		object.uM12155mmHMC,    
		object.uUSBofors40mmAA, 
		object.uM2A1105mmArt,   
		object.uM1155mmArt,     
		object.uUSTruck,        
		object.uBritishVeteran,
		object.uBritishRegular,
		object.uBritishPIAT,   
		object.uBritishMG,     
		object.uCommandos, 
		object.u6pdrAT,			
		object.uDaimlerAC,     
		object.uBritishMotorTroops,  
		object.uShermanIIaBde,
		object.uShermanIIa,    
		object.uChurchillMkIV, 
		object.uShermanVCFirefly,
		object.uGurkhas,       
		object.uBishop25pdr,   
		object.uUKBofors40mmAA,
		object.u25pdrMKIIArt,  
		object.u55inchArt,     
		object.uUKTruck, 
		object.uCdnMotInf,
		object.uCdnShermanIII,
		object.uCdnM7105mm,
		object.uCdnMotorTroops,
		object.uCdnInfantry,
		object.uRadarStation,
		object.uSAMotInf,
		object.uSAShermanIII,
		object.uSASexton,
		object.uSAMotorTroops,
		object.uBraz1stInf,
		object.uGurkhas,
		object.uIndianInfantry,
		object.u4thIndianDiv,
		object.uIndian25pdrArt,
		object.uPolishInfantry,
		object.uPolishShermanIII,
		object.uPolish25pdrArt,
		object.uInfanterieColoniale,
		object.uGoumierMarocain,
		object.uTirailleurAlgerien,
		object.uFreeFrenchInfantry,
		object.uFreeFrench105Art,
		object.uFreeFrenchM10,
		object.uNZMotInf,
		object.uNZShermanIII,
		object.uNZ25pdrArt,
		object.uChurchillAVRE,
		object.uUS10thMountain,
		object.uUS1stInfantry,
		object.uCromwellMkVII,
		object.uM4A1ShermanInd,
		object.uGreekMtnBde,
		object.uJewishBde,
		object.uDocks,
		object.uDocks,			
		object.uSupplyHub,		
		object.uBritish52ndMtn,
	},
    unloadFailureMessage = "%STRING1 units can only disembark units within a city or from a beachhead tile."
}

-- Leander Class CL can only unload from Beachhead tile
local function leanderClassCLUnload(leanderClassCL)
    if gen.unitTypeOnTile(leanderClassCL.location, object.uBeachhead) then
        return true
    else
        return false
    end
end

beachShipSettings[object.uLeanderClassCL.id] = {
    unloadCheckFunction = leanderClassCLUnload, --
	restrictBeachUnload = false,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = false,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
	-- Leander Class CL can only carry Allied Generals
	forbidBoarding = {object.uEngineer,
		object.uGIVeteran,
		object.uGIRegular,
		object.uUSBazooka,
		object.uRanger, 
		object.uDUKW,     
		object.uArmoredInfantry,
		object.uM5Stuart,       
		object.uM4A1Sherman,    
		object.uM4A376WSherman, 
		object.uM10Wolverine, 
		object.uRanger,
		object.u82ndAirborne,
		object.uM157mmAT,		
		object.uM7105mmHMC,     
		object.uM12155mmHMC,    
		object.uUSBofors40mmAA, 
		object.uM2A1105mmArt,   
		object.uM1155mmArt,     
		object.uUSTruck,        
		object.uBritishVeteran,
		object.uBritishRegular,
		object.uBritishPIAT,   
		object.uBritishMG,     
		object.uCommandos,
		object.u1stAirborne,		
		object.u6pdrAT,	
		object.uDaimlerAC,     
		object.uBritishMotorTroops,  
		object.uShermanIIaBde,
		object.uShermanIIa,    
		object.uChurchillMkIV, 
		object.uShermanVCFirefly,
		object.uGurkhas,       
		object.uBishop25pdr,   
		object.uUKBofors40mmAA,
		object.u25pdrMKIIArt,  
		object.u55inchArt,     
		object.uUKTruck, 
		object.uCdnMotInf,
		object.uCdnShermanIII,
		object.uCdnM7105mm,
		object.uCdnMotorTroops,
		object.uCdnInfantry,
		object.uRadarStation,
		object.uSAMotInf,
		object.uSAShermanIII,
		object.uSASexton,
		object.uSAMotorTroops,
		object.uBraz1stInf,
		object.uGurkhas,
		object.uIndianInfantry,
		object.u4thIndianDiv,
		object.uIndian25pdrArt,
		object.uPolishInfantry,
		object.uPolishShermanIII,
		object.uPolish25pdrArt,
		object.uInfanterieColoniale,
		object.uGoumierMarocain,
		object.uTirailleurAlgerien,
		object.uFreeFrenchInfantry,
		object.uFreeFrench105Art,
		object.uFreeFrenchM10,
		object.uNZMotInf,
		object.uNZShermanIII,
		object.uNZ25pdrArt,
		object.uChurchillAVRE,
		object.uUS10thMountain,
		object.uUS1stInfantry,
		object.uCromwellMkVII,
		object.uM4A1ShermanInd,
		object.uGreekMtnBde,
		object.uJewishBde,	
		object.uDocks,	
		object.uSupplyHub,		
		object.uBritish52ndMtn,
	},
    unloadFailureMessage = "%STRING1 units can only disembark units within a city or from a beachhead tile."
}

-- -- Leander Class CL can only unload from Beachhead tile
-- local function leanderClassCLUnload(leanderClassCL)
    -- if gen.unitTypeOnTile(leanderClassCL.location, object.uBeachhead) then
        -- return true
    -- else
        -- return false
    -- end
-- end

-- beachShipSettings[object.uLeanderClassCL.id] = {
    -- unloadCheckFunction = leanderClassCLUnload,
-- }


beachShipSettings[object.uLandingCraft.id] = {
    restrictBeachUnload = false,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = false,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
	-- Landing Craft can carry all types of ground units
	forbidBoarding = {object.uDUKW,     
		object.uUSTruck,        
		object.uUKTruck,
		object.uDocks,	
		object.uSupplyHub,		
	},
    unloadFailureMessage = "%STRING1 units can disembark units within a city or beach tile."
}

beachShipSettings[object.uLibertyShip.id] = {
    restrictBeachUnload = true,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = true,
    forbidPeacePortBoarding = true,
	forbidBoarding = {object.uDocks,		
	},
    unloadFailureMessage = "%STRING1 units can only disembark units within a city or from a beachhead tile."
}

-- Liberty Ship can only unload from Beachhead tile
local function libertyShipUnload(libertyShip)
    if gen.unitTypeOnTile(libertyShip.location, object.uBeachhead) then
        return true
    else
        return false
    end
end

beachShipSettings[object.uLibertyShip.id] = {
    unloadCheckFunction = libertyShipUnload,
}



--[[
-- Here are some testing examples
beachShipSettings[gen.original.uTrireme.id] = {
    restrictBeachUnload = true,
    restrictPeacePortUnload = true,
    restrictAlliedPortUnload = true,
    forbidBeachBoarding = true,
    forbidPeacePortBoarding = true,
    forbidAlliedPortBoarding = true,
    unloadFailureMessage = "%STRING1 units can only disembark units within a city."
}

beachShipSettings[gen.original.uMechInf.id] = {
    restrictBeachUnload = true,
    restrictPeacePortUnload = true,
    forbidBeachBoarding = true,
    unloadFailureMessage = "%STRING1 units can only disembark within a city, or beside an allied city.",
}

beachShipSettings[gen.original.uGalleon.id] = {
    restrictBeachUnload = 0,
    restrictPeacePortUnload = 2,
    restrictAlliedPortUnload = 3,
    beachUnloadPenalty = 2,
    unloadFailureMessage = "%STRING1 units can only disembark units onto a beach if they haven't moved, can only disembark units beside a city with whom they are at peace if they have not moved more than 2 spaces, and can't disembark units beside allied cities unless they haven't moved more than 3 spaces."
}

--beachShipSettings[gen.original.uArmor.id] = {
--    restrictBeachUnload = -3,
--    restrictPeacePortUnload = -6,
--    beachUnloadPenalty = -4,
--}

beachShipSettings[gen.original.uArchers.id] = {
    boardingCheckFunction = function(boardingUnit,shipType)
        return shipType == gen.original.uCaravel 
    end
    -- archers can only board a caravel outside of port
}

beachShipSettings[gen.original.uArmor.id] = {
    restrictBeachUnload = -3,
    restrictPeacePortUnload = -6,
    beachUnloadPenalty = function(unit) unit.damage = unit.damage+5 end,
    -- note, repeatedly activating the unit will cause more damage each time
    -- with this setup unit.damage = math.max(unit.damage,5) would be an alternative,
    -- as would keeping track of units already activated this turn
}

beachShipSettings[gen.original.uLegion.id] = {
    forbidBoarding = {gen.original.uFrigate},
}
beachShipSettings[gen.original.uFrigate.id] = {
    forbidBoarding = {gen.original.uMusketeers}
}
--]]



-- Carrier Units
-- If a unit is not within the carrierSettings table, it will either always
-- be a carrier, or never be a carrier
-- carrierSettings[unitType.id] = table with keys:
--  canNotUse = nil or table of unit types
--      units in this table can not land on the specified carrier
--      (if the unit is land/sea, this means the carrier does not have
--      carrier status when that unit is activated)
--  setPropertyFor = nil or table of tribes
--      the carrier property is always enabled for these tribes
--      (You can avoid air protected stacks for carriers by using
--      this and assigning a different carrier unit type to each tribe/alliance)
--      (note that range 0 or 1 units don't provide air stack anyway)
--  clearPropertyFor = nil or table of tribes
--      the carrier property is always disabled for these tribes
--      (not sure why you'd want this)


local carrierSettings = {}

-- UNDER CARRIER SETTINGS THE HMS Illustrious CAN ONLY CARRY the BRITISH Seafire MkII AND Fairey Barracuda AIRCRAFT
carrierSettings[object.uIllustriousClassCV.id] = {
    -- canNotUse = {gen.original.uBomber, gen.original.uStlthBmbr, gen.original.uNuclearMsl},
	canNotUse = {object.uBeaufighterMkVI, object.uKittyhawk, object.uSpitfireMkIX, 
	object.uHurricaneIIc, object.uMartinBaltimore, object.uVickersWellington, object.uBostonIII, object.uSunderlandIII, 
	object.uP40Warhawk, object.uP38Lightning, object.uP51DMustang, 
	object.uA36Apache, object.uB24Liberator, object.uB25Mitchell, object.uB26Marauder, object.uB17Fortress},
	hold = 2,
}

-- carrierSettings[object.uWacoGlider.id] = {
    -- canNotUse = {object.uGIVeteran, object.uP40Warhawk, object.u105mmShell, object.u155mmShell },
-- }

--[[
carrierSettings[gen.original.uCarrier.id] = {
    canNotUse = {gen.original.uBomber, gen.original.uStlthBmbr, gen.original.uNuclearMsl},
    setPropertyFor = {civ.getTribe(2),civ.getTribe(3)},
    clearPropertyFor = {civ.getTribe(4)},
}
carrierSettings[gen.original.uTrireme.id] = {
    canNotUse = {gen.original.uBomber, gen.original.uStlthBmbr, gen.original.uStlthFtr},
}
--]]



navy.registerBeachShipSettings(beachShipSettings)
navy.registerCarrierSettings(carrierSettings)
-- uncomment this line if you want these settings to apply to the AI
-- This is not tested, some parts might not apply to the AI anyway
-- (I think the AI will still be able to land on beaches as normal),
-- and the AI might not understand how it is supposed to do things
-- (it might have difficulty with not being able to load from a beach)
-- navy.applySettingsToAI()

local navySettings = {}
gen.versionFunctions(navySettings,versionNumber,fileModified,"MechanicsFiles".."\\".."navySettings.lua")
return navySettings


