--[[
supplyLines.lua
v1.5
by Knighttime

TERMINOLOGY:
	A "supply depot" is a tile that meets the specified condition(s) to be able to provide supplies (most commonly, to a unit).
	A "supply line" is a set of tiles that meets the specified condition(s) and leads from an object needing supplies (most commonly, a unit)
		to a supply depot. If there are multiple such paths, as is common, this module will identify one such that no *shorter* one exists.
		(No indication will be provided as to whether additional valid paths exist or how their lengths compare.)
	A "supply link" is a tile that meets the specified condition(s) and is eligible to serve as part of a supply line.
	If no supply depot can be found (that is, no supply line exists), the "supply zone" is the set of all supply links.
		If any one of these contained a supply depot, a supply line could be formed.
		Note that depending on the options selected, a port city added *adjacent to* the supply zone may also become a valid supply depot.

SCOPE:
	Supply lines can never cross between maps (obviously this is only relevant within games using multiple maps). In other words, both the
		object needing supplies and the supply depot providing supplies must exist on the same map, in order for a supply line to be formed.
	Supply line lengths are measured in *tiles*, not "movement points" or "turns for unit X to travel".
		In other words, while base terrain types and/or the presence of roads can be used in the criteria for supply links, the movement
		cost of terrain and/or the benefits of roads/railroads are not relevant when measuring the supply line length.
		Furthermore, a diagonal connection between tiles that only share a corner still counts as a distance of 1 tile, just like a connection
		between tiles that share an edge.
	Enemy units will not exert a Zone of Control. A supply line may not pass through a tile containing an enemy unit (that is, such a tile can
		never be a valid supply link) but the supply line may pass through tiles directly adjacent to enemy units. This may be added as a
		configurable option to a future version of this module.

INSTRUCTIONS:
	1. Edit the constants immediately below as you wish (beginning at line 64)
	2. Edit the functions, or the conditions within them, in "isValidSupplyDepot" below (beginning at line 94)
	3. Edit the functions, or the conditions within them, in "isValidSupplyLink" below (beginning at line 156)
	4. Edit the function "statusCheckNeeded" below (beginning at line 196)
	5. Integrate this module with your other events (see examples in events.lua, provided alongside this file, for details)
		Note that if you are using the Lua Scenario Template (created by Prof. Garfield) for your other events, the integration points will not be placed in a file named "events.lua".
		Follow the instructions and documentation for that Template in order to place each piece of the integration code in the correct location.
		
ADDITIONAL AVAILABLE FUNCTIONS:
	The following "helper" functions are available within supplyLinesCore and can be called from this file:
		supplyLinesCore.isAlly
		supplyLinesCore.isEnemy
		supplyLinesCore.containsCity
		supplyLinesCore.containsAlliedTribeCity
		supplyLinesCore.containsEnemyTribeCity
		supplyLinesCore.containsOtherTribeCity
		supplyLinesCore.containsYourTribeCity
		supplyLinesCore.containsUnit
		supplyLinesCore.containsAlliedTribeUnit
		supplyLinesCore.containsEnemyTribeUnit
		supplyLinesCore.containsOtherTribeUnit
		supplyLinesCore.containsYourTribeUnit
		supplyLinesCore.containsUnitType
		supplyLinesCore.containsAlliedTribeUnitType
		supplyLinesCore.containsEnemyTribeUnitType
		supplyLinesCore.containsOtherTribeUnitType
		supplyLinesCore.containsYourTribeUnitType
	These may be very useful in the criteria for "isValidSupplyDepot" and "isValidSupplyLink".
	Some examples using them are provided in those sections below.
]]

local supplyLinesCore = require("supplyLinesCore")
local supplyLines = { }
supplyLines.supplyLineLastUnitId = -1

-- ===== EDITABLE CONSTANTS TO ADJUST SUPPLY LINE MODULE BEHAVIOR ====
-- Note: If you are using the Lua Scenario Template and wish to reference entries in the 'object' table in this file,
--		 uncomment the following line:
local object = require("object")

supplyLines.MAX_SUPPLY_LINE_LENGTH = 12							-- A positive number should be self-explanatory. Zero is permitted and means a unit must be *on* a supply depot in order
																--		to be supplied. Any negative number means "infinite", that is, a supply line has no length limit and can extend to
																--		any point on the map. Note that this defines the maximum supply line length that will ever be analyzed. However, it is
																--		possible to define specific supply depot criteria that require a *shorter* supply line length, for that type of depot.
																--		(An example of this is provided on line 115.) Therefore, this should always be the *longest* possible supply line ever
																--		permitted for any type of depot.
supplyLines.ONLY_COASTAL_CITIES_SUPPLY_ACROSS_OCEAN = false		-- Limits behavior, *if* ocean tiles are permitted within isValidSupplyLink. If they are not, this has no effect.
																-- 		Essentially provides a hybrid option between two extremes: (1) blocking supply lines from using ocean tiles at all,
																--		or (2) treating them exactly the same as every other base terrain type.
supplyLines.SHOW_CONSOLE_MESSAGES = false						-- A value of true will result in informational messages being displayed to the console, which can be useful during the initial
																--		implementation and testing. However, this has a slight performance impact, so it is recommended to set this to false
																--		before the final release of your scenario.
supplyLines.SHOW_SUPPLY_LINE_KEYCODE = nil						-- Keycode you wish to use to trigger supply line review, intended for scenario design and testing.
																-- 		It is strongly recommended to set this to "nil" in order to disable this functionality, for an official scenario release.
supplyLines.SHOW_SUPPLY_LINE_KEYLABEL = ""						-- Label of this keycode (may be useful for dialog boxes)
supplyLines.SUPPLY_MARKER_UNITTYPE = nil						-- It is recommended that you create a unit type named "Supply Marker" in Rules.txt, with an appropriate graphic in Units.bmp,
																--		and then find or reference that unit here. If you do not wish to utilize the supply line review feature, this may be
																--		set to nil.
supplyLines.SUPPLY_MARKER_MAX_QUANTITY = 1000					-- Supply markers are created as temporary units, so a high number may exceed the limit on the number of units permitted.
																-- 		Recommended max is 1000 which supports a supplyLines.MAX_SUPPLY_LINE_LENGTH of 15
																-- 		Potential number of supply markers is: ((supplyLines.MAX_SUPPLY_LINE_LENGTH*2)+1)^2
																--		Within TOTPP, the default max number of units is 4096, but this is configurable.
																--		Large values for this constant may result in poor performance.
-- ===================================================================

supplyLines.isValidSupplyDepot = { }
--[[
NOTE: "isValidSupplyDepot" is a table of functions. Some predefined functions are shown below, but it is also permissible for you to add
		new functions into this table, following the same pattern as the ones shown below.
	  You could then return these new functions from the "statusCheckNeeded" function, and they will be run at the appropriate times.
	  Each function within "isValidSupplyDepot" must define whether or not a tile meets the	criteria to serve as a supply depot.
]]
supplyLines.isValidSupplyDepot.main = function (tile, currentSupplyLineLength, forTribe, supplyDepotTileSet)
	local isDepot = false
	local description = ""
	if
-- ===== EDITABLE CONDITIONS FOR A TILE TO SERVE AS A SUPPLY DEPOT ====
--	Tile contains one of your cities:
--		(supplyLinesCore.containsYourTribeCity(tile, forTribe) == true)
--	Tile contains one of your cities and that city has a Barracks:
--		(supplyLinesCore.containsYourTribeCity(tile, forTribe) == true and civ.hasImprovement(tile.city, civ.getImprovement(2)) == true)
--	Tile contains one of your cities and that city has a Supply Depot (Recycling Center):
		(supplyLinesCore.containsYourTribeCity(tile, forTribe) == true and civ.hasImprovement(tile.city, object.iSupplyDepot) == true)
--	Tile contains one of your units of a specific type:
--		(supplyLinesCore.containsYourTribeUnitType(tile, civ.getUnitType(0), forTribe) == true)
--	Tile contains one of your units of a specific type, within a shorter distance than the maximum supply line length:
	or	(supplyLinesCore.containsYourTribeUnitType(tile, object.uSupplyHub, forTribe) == true and currentSupplyLineLength <= 8)
--	Tile contains an airbase and one of your units:
--	or	(tile.improvements & 0x42 == 0x42 and supplyLinesCore.containsYourTribeUnit(tile, forTribe) == true)
-- ====================================================================
	then
		isDepot = true
		description = "valid supply depot at " .. tile.x .. "," .. tile.y .. "," .. tile.z
		if tile.city ~= nil then
			description = description .. " (city of " .. tile.city.name .. ")"
		end
	end
	return isDepot, description
end
supplyLines.isValidSupplyDepot.tileSet = function (tile, currentSupplyLineLength, forTribe, supplyDepotTileSet)
	local isDepot = false
	local description = ""
-- ===== THIS EXAMPLE DOES NOT CONTAIN EDITABLE CONDITIONS ====
--	Instead, it illustrates usage of an extra parameter, "supplyDepotTileSet"
--	This allows you to calculate or define a table of tiles *in your calling program* that serve as valid supply depots,
--		and then pass that table in as a parameter.
--	This example then checks to see whether the tile at hand is within that set.
--	This could potentially allow you to "daisy-chain" supply depots of various types together, with specific rules, by making multiple calls to "supplyLines.hasSupplyLine()"
--		and accumulating the results.
-- ======================================================================
	for _, depotTile in ipairs(supplyDepotTileSet) do
		if depotTile == tile then
			isDepot = true
			description = "valid supply depot at " .. tile.x .. "," .. tile.y .. "," .. tile.z .. " (within supplyDepotTileSet)"
			break
		end
	end
	return isDepot, description
end
--[[
	Right now, the two additional functions below simply call the first one defined above.
	But if you wanted to have separate and unique conditions for air units vs. land units, you could instead provide complete definitions
		for them, with the appropriate logic.
]]
supplyLines.isValidSupplyDepot.land = supplyLines.isValidSupplyDepot.main
supplyLines.isValidSupplyDepot.air = supplyLines.isValidSupplyDepot.main

supplyLines.isValidSupplyLink = { }
--[[
NOTE: "isValidSupplyLink" is a table of functions. Some predefined functions are shown below, but it is also permissible for you to add
		new functions into this table, following the same pattern as the ones shown below.
	  You could then return these new functions from the "statusCheckNeeded" function, and they will be run at the appropriate times.
	  Each function within "isValidSupplyLink" must define whether or not a tile meets thecriteria to serve as a supply link.
]]
supplyLines.isValidSupplyLink.main = function (tile, forTribe)
	local isLink = false
	if
-- ===== EDITABLE CONDITIONS FOR SUPPLY LINK ====
--	Tile may not contain a city belonging to a different tribe:
		(supplyLinesCore.containsOtherTribeCity(tile, forTribe) == false)
--	Tile may not contain a city belonging to barbarians or a tribe with whom you are at war:
--		(supplyLinesCore.containsEnemyTribeCity(tile, forTribe) == false)
--	Tile may not be occupied by a unit belonging to a different tribe:
	and	(supplyLinesCore.containsOtherTribeUnit(tile, forTribe) == false)
--	Tile may not be occupied by a unit belonging to barbarians or a tribe with whom you are at war:
--		(supplyLinesCore.containsEnemyTribeUnit(tile, forTribe) == false)
--	Tile may not be specific terrain types:
	and	(tile.baseTerrain ~= object.bSwamp)
--	and	(tile.baseTerrain.type ~= 10)	-- Ocean
--	Tile must contain road or railroad:
--	and	(tile.improvements & 0x10 == 0x10)
--	Tile of a certain terrain type must contain road or railroad:
	and	(tile.baseTerrain ~= object.bApennines or tile.improvements & 0x10 == 0x10)
-- ===============================================
	then
		isLink = true
	end
	return isLink
end
--[[
	Right now, the two additional functions below simply call the first one defined above.
	But if you wanted to have separate and unique conditions for air units vs. land units, you could instead provide complete definitions
		for them, with the appropriate logic.
]]
supplyLines.isValidSupplyLink.land = supplyLines.isValidSupplyLink.main
supplyLines.isValidSupplyLink.air = supplyLines.isValidSupplyLink.main

--[[
This function allows you to define which game objects (cities or units) should have their supplied status checked.
This function should return true if that game object's supply status matters in any way. If the supply status of the object
	is irrelevant, the function should return false.
]]
supplyLines.statusCheckNeeded = function (gameObject)
	local city = nil
	local unit = nil
	if civ.isCity(gameObject) then
		city = gameObject
	elseif civ.isUnit(gameObject) then
		unit = gameObject
	end

-- ===== EDITABLE CONDITIONS FOR A UNIT TO HAVE ITS SUPPLY STATUS CHECKED ====
	if unit ~= nil and
		unit.owner.isHuman == true and
		unit.owner == object.pAllies and
		unit.location.baseTerrain.type ~= 10 and
		-- unit.location.baseTerrain.type ~= 9 and
		unit.location ~= civ.getTile(129,185,0) and -- For tile that tranported units are temporarily placed in
		--(unit.type.domain == 0 or unit.type.domain == 1) and
		unit.type ~= object.uEngineer and
		unit.type ~= object.uUKTruck and
		unit.type ~= object.uUSTruck and
		unit.type ~= object.uDUKW and
		unit.type ~= object.uCommandos and
		unit.type ~= object.u1stAirborne and
		unit.type ~= object.u82ndAirborne and
		unit.type ~= object.uRanger and
		unit.type ~= object.uPartigiani and
		unit.type ~= object.uGenPatton and
		unit.type ~= object.uGenClark and
		unit.type ~= object.uGenTruscott and
		unit.type ~= object.uGenMontgomery and
		unit.type ~= object.uGenLeese and
		unit.type ~= object.uGenMcCreery and
		unit.type ~= object.uGreekMtnBde and
		unit.type ~= object.uGoumierMarocain and
		unit.type ~= object.uUS10thMountain and
		unit.type ~= object.uBritish52ndMtn -- and
		--unit.type ~= object.uSupplyHub and
		--unit.type ~= object.uM1155mmArt
	then
		if unit.type.domain == 0 then
			return true, supplyLines.isValidSupplyDepot.land, supplyLines.isValidSupplyLink.land
		-- elseif unit.type.domain == 1 then
			-- return true, supplyLines.isValidSupplyDepot.air, supplyLines.isValidSupplyLink.air
		end
-- ===========================================================================

	else
		return false, nil, nil
	end
end

--[[
The two functions below provide entry points into the module from your scenario.
They should not need to be edited under normal circumstances.
]]
supplyLines.getSupplyStatus = function (gameObject)
	local gameObjectTypeString = ""
	local gameObjectName = ""
	if civ.isCity(gameObject) then
		gameObjectTypeString = "city"
		gameObjectName = gameObject.name
	elseif civ.isUnit(gameObject) then
		gameObjectTypeString = "unit"
		gameObjectName = gameObject.type.name .. " (domain " .. gameObject.type.domain .. ")"
		supplyLines.supplyLineLastUnitId = gameObject.id
	end
	if gameObjectTypeString ~= "" then
		if supplyLines.SHOW_CONSOLE_MESSAGES then
			print("Detected call to supplyLines.getSupplyStatus() for " .. gameObjectTypeString .. " " .. gameObject.id .. " (" .. gameObjectName .. ", " .. gameObject.owner.adjective .. ") at " .. gameObject.x .. "," .. gameObject.y .. "," .. gameObject.z)
		end
		local supplyStatus = nil
		local checkNeeded, depotFunction, linkFunction = supplyLines.statusCheckNeeded(gameObject)
		if checkNeeded then
			supplyStatus = supplyLinesCore.hasSupplyLine(supplyLines, gameObject, depotFunction, nil, linkFunction, 0)
		end
		return supplyStatus
	else
		civ.ui.text("Supply Lines Module ERROR! Detected call to supplyLines.getSupplyStatus() but object was not a city or a unit!")
	end
end

supplyLines.showSupplyLine = function ()
	supplyLinesCore.showSupplyLine(supplyLines)
end

print("supplyLines.lua parsed successfully at " .. os.date("%c"))
print("    Lines found: " .. debug.getinfo(1).currentline + 8)

return supplyLines
