--
local versionNumber = 1
local fileModified = true -- set this to true if you change this file for your scenario
-- if another file requires this file, it checks the version number to ensure that the
-- version is recent enough to have all the expected functionality
-- if you set fileModified to true, the error generated if this file is out of date will
-- warn you that you've modified this file
--
--
--
--
--  This file allows for a scenario designer to organize events
--  in a manner similar to the legacy system, in that you don't have to
--  group all events of the same type in the same place.
--  This may be more convenient in some cases.
--
--  You can create events in multiple files, but event order is only
--  guaranteed to be preserved within files.  That is, for two files
--  and 4 events with the same execution point
--      discreteEventsFile1.lua
--          Event A
--          Event B
--      discreteEventsFile2.lua
--          Event Y
--          Event Z
--  A will be checked before B and Y before Z,
--  but both A, B, Y, Z and Y, Z, A, B
--  are possible orders to check and execute the code
--
--
--
--

-- ===============================================================================
--
--          Require Lines etc.
--
-- ===============================================================================
-- This section is for the 'require' lines for this file, and anything
-- else that must be at the top of the file.


---@module "discreteEventsRegistrar"
local discreteEvents = require("discreteEventsRegistrar"):minVersion(4)
---@module "generalLibrary"
local gen = require("generalLibrary"):minVersion(1)
---@module "param"
local param = require("parameters"):minVersion(1)
local object = require("object")
---@module "text"
local text = require("text"):minVersion(1)
---@module "diplomacy"
local diplomacy = require("diplomacy"):minVersion(1)
---@module "delayedAction"
local delayed = require("delayedAction"):minVersion(1)
local calendar = require("calendar")
local keyboard = require("keyboard")
local flag = require("flag")
local counter = require("counter")
local tables = require("tables")

-- local unitData = require("unitData")
-- unitData.defineFlag("stackPenaltyNotApplied",true,"onTribeTurnEnd")

local func = require("functions")
local onTurn = require("onTurn")

local civlua = require("civluaModified")
local tileData = require("tileData")

local function round (decimal)
	-- Rounds a numeric value to the nearest integer and returns the integer
	return math.floor(decimal + 0.5)
end

-- ===============================================================================
--
--          Discrete Events
--
-- ===============================================================================

---&autoDoc onActivateUnit
--Registers a function to be called every time a unit is activated. The callback takes the unit activated as a parameter, and the source of unit activation. `source` is `true` if activated by keyboard or mouse click, `false` if activated by the game itself. `repeatMove` is `true` if it's a repeat activation caused by moving, `false` otherwise.
--If the function returns true, the unit activation will be cancelled.
--No further activation code will be executed, and the unit's type
--will temporarily be set to have 0 movement points.
--If the function returns function(unit), then the unit activation
--will be cancelled, and the function returned will be executed
--just before another unit is activated.  (You may wish to put
--the unit to sleep, for example.)
--Not returning anything is equivalent to returning nil, which is
--acceptable, and keeps the unit activation going.
discreteEvents.onActivateUnit(function(unit,source,repeatMove)
    -- if _global.eventTesting then
        -- civ.ui.text("Unit activation Discrete Event")
    -- end

    -- =============== -- (by Prof. Garfield)
	-- EXTRA CODE TO WORK WITH 'LANDING CRAFT OR LIBERTY SHIPS' section just below
	gen.getState().convertedBeachTiles = gen.getState().convertedBeachTiles or {} -- initialize if not already
    --convetedBeachTiles[gen.getTileID(normallyBeachTile)] = true
    --Beach tiles are converted to riverCrossing tiles to prevent units that
    --are not in landing craft from unloading onto them, since they
    --should only unload onto beaches with the 'town' resource.
    local convertedBeachTiles = gen.getState().convertedBeachTiles
    -- undo any converted beach tiles:
    for tileID,_ in pairs(convertedBeachTiles) do
        local tile = gen.getTileFromID(tileID)
        tile.baseTerrain = object.bBeach
    end 
	-----------------

    -- ALLIED GROUND UNITS ON LANDING CRAFT OR LIBERTY SHIPS CAN ONLY LAND ON BEACH TILES
	if unit.owner == object.pAllies 
	and unit.type.domain == 0 and unit.location.baseTerrain == object.bOcean 
    and ((gen.unitTypeOnTile(unit.location,object.uLandingCraft)) or (gen.unitTypeOnTile(unit.location,object.uBeachhead))) then
        -- Temporarily change terrain 'impassable' flag to prevent infantry units from landing on non-beach tiles
		object.bRough.impassable = true
		object.bTemperateGrassland.impassable = true
        object.bForest.impassable = true
		object.bHills.impassable = true
		object.bGrassland.impassable = true
		object.bSwamp.impassable = true
		object.bTownCenter.impassable = true
		object.bTemperateHills.impassable = true
		object.bTemperateRough.impassable = true
		object.bRiverCrossing.impassable = true

        -- ==== (by Prof. Garfield)
		if (not unit.carriedBy) or (unit.carriedBy.type ~= object.uLandingCraft) then
            for _,tile in pairs(gen.getAdjacentTiles(unit.location)) do
                if tile.baseTerrain == object.bBeach and tile.terrain ~= object.tTownBeach then
                    convertedBeachTiles[gen.getTileID(tile)] = true
                    tile.baseTerrain = object.bRiverCrossing
                end
            end
        end
		-- ====
		
		-- Temporarily remove infantry * units 'impassable' flag to prevent them from landing on non-beach tiles
		-- Exception 'Commandos' and 'Rangers' can disembark on non-beach tiles
		-- Note *: the impassable terrain override flag for Mechanized units is already not selected in the rules.txt @UNITS_ADVANCED section
		gen.removeCanCrossImpassableTerrain(object.uEngineer) 
		gen.removeCanCrossImpassableTerrain(object.uBritishVeteran) 
		gen.removeCanCrossImpassableTerrain(object.uGIVeteran) 
		gen.removeCanCrossImpassableTerrain(object.uGenPatton) 
		gen.removeCanCrossImpassableTerrain(object.uGenClark) 
		gen.removeCanCrossImpassableTerrain(object.uGenTruscott) 
		-- gen.removeCanCrossImpassableTerrain(object.uRanger) 
		gen.removeCanCrossImpassableTerrain(object.u82ndAirborne) 
		gen.removeCanCrossImpassableTerrain(object.uGenMontgomery) 
		gen.removeCanCrossImpassableTerrain(object.uGenLeese) 
		gen.removeCanCrossImpassableTerrain(object.uGenMcCreery) 
		gen.removeCanCrossImpassableTerrain(object.uBritishPIAT) 
		gen.removeCanCrossImpassableTerrain(object.uBritishMG) 
		-- gen.removeCanCrossImpassableTerrain(object.uCommandos) 
		gen.removeCanCrossImpassableTerrain(object.u1stAirborne) 
		gen.removeCanCrossImpassableTerrain(object.uCdnInfantry) 
		gen.removeCanCrossImpassableTerrain(object.uBraz1stInf) 
		gen.removeCanCrossImpassableTerrain(object.uGurkhas) 
		gen.removeCanCrossImpassableTerrain(object.uIndianInfantry) 
		gen.removeCanCrossImpassableTerrain(object.u4thIndianDiv) 
		gen.removeCanCrossImpassableTerrain(object.uPolishInfantry) 
		gen.removeCanCrossImpassableTerrain(object.uInfanterieColoniale) 
		gen.removeCanCrossImpassableTerrain(object.uGoumierMarocain) 
		gen.removeCanCrossImpassableTerrain(object.uTirailleurAlgerien) 
		gen.removeCanCrossImpassableTerrain(object.uGreekMtnBde) 
		gen.removeCanCrossImpassableTerrain(object.uJewishBde) 
		gen.removeCanCrossImpassableTerrain(object.uGIRegular) 
		gen.removeCanCrossImpassableTerrain(object.uUS1stInfantry) 
		gen.removeCanCrossImpassableTerrain(object.uUS10thMountain) 
		gen.removeCanCrossImpassableTerrain(object.uUSBazooka) 
		gen.removeCanCrossImpassableTerrain(object.uBritishRegular) 
		
    else
		-- Restore terrain 'impassable' flag to false if not trying to land ground unit from naval transport 
		object.bRough.impassable = false
		object.bTemperateGrassland.impassable = false
        object.bForest.impassable = false
		object.bHills.impassable = false
		object.bGrassland.impassable = false
		object.bSwamp.impassable = false
		object.bTownCenter.impassable = false
		object.bTemperateHills.impassable = false
		object.bTemperateRough.impassable = false
		object.bRiverCrossing.impassable = false
		
		-- Restore infantry units 'impassable' flag if not trying to land ground unit from naval transport 
		gen.giveCanCrossImpassableTerrain(object.uEngineer) 
		gen.giveCanCrossImpassableTerrain(object.uBritishVeteran) 
		gen.giveCanCrossImpassableTerrain(object.uGIVeteran) 
		gen.giveCanCrossImpassableTerrain(object.uGenPatton) 
		gen.giveCanCrossImpassableTerrain(object.uGenClark) 
		gen.giveCanCrossImpassableTerrain(object.uGenTruscott) 
		gen.giveCanCrossImpassableTerrain(object.uRanger) 
		gen.giveCanCrossImpassableTerrain(object.u82ndAirborne) 
		gen.giveCanCrossImpassableTerrain(object.uGenMontgomery) 
		gen.giveCanCrossImpassableTerrain(object.uGenLeese) 
		gen.giveCanCrossImpassableTerrain(object.uGenMcCreery) 
		gen.giveCanCrossImpassableTerrain(object.uBritishPIAT) 
		gen.giveCanCrossImpassableTerrain(object.uBritishMG) 
		gen.giveCanCrossImpassableTerrain(object.uCommandos) 
		gen.giveCanCrossImpassableTerrain(object.u1stAirborne) 
		gen.giveCanCrossImpassableTerrain(object.uCdnInfantry) 
		gen.giveCanCrossImpassableTerrain(object.uBraz1stInf) 
		gen.giveCanCrossImpassableTerrain(object.uGurkhas) 
		gen.giveCanCrossImpassableTerrain(object.uIndianInfantry) 
		gen.giveCanCrossImpassableTerrain(object.u4thIndianDiv) 
		gen.giveCanCrossImpassableTerrain(object.uPolishInfantry) 
		gen.giveCanCrossImpassableTerrain(object.uInfanterieColoniale) 
		gen.giveCanCrossImpassableTerrain(object.uGoumierMarocain) 
		gen.giveCanCrossImpassableTerrain(object.uTirailleurAlgerien) 
		gen.giveCanCrossImpassableTerrain(object.uGreekMtnBde) 
		gen.giveCanCrossImpassableTerrain(object.uJewishBde) 
		gen.giveCanCrossImpassableTerrain(object.uGIRegular) 
		gen.giveCanCrossImpassableTerrain(object.uUS1stInfantry) 
		gen.giveCanCrossImpassableTerrain(object.uUS10thMountain) 
		gen.giveCanCrossImpassableTerrain(object.uUSBazooka) 
		gen.giveCanCrossImpassableTerrain(object.uBritishRegular) 		
    end	

end)
---&endAutoDoc

---&autoDoc onCityDestroyed
discreteEvents.onCityDestroyed(function(city) 
    if _global.eventTesting then
        civ.ui.text("City destroyed discrete event test")
    end
end)
---&endAutoDoc

---&autoDoc onBribeUnit
discreteEvents.onBribeUnit(function(unit,previousOwner)
    if _global.eventTesting then
        civ.ui.text("Bribe unit discrete event test")
    end
end)
---&endAutoDoc

---&autoDoc onCityFounded
discreteEvents.onCityFounded(function(city) 
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onCityFounded for "..city.name)
    end
    -- the cityCancelled() function is executed if the player
    -- decides not to found the city after all
    -- (so you can undo terrain changes, etc.
    local function cityCancelled()
        if _global.eventTesting then
            civ.ui.text("discreteEvents.onCityFounded city cancelled for "..city.name)
        end
    end
    return cityCancelled
end)
---&endAutoDoc

---&autoDoc onCityProcessed
discreteEvents.onCityProcessed(function(city) 
    --civ.ui.text("City processed discrete event test for city "..city.name)

end)
---&endAutoDoc

---&autoDoc onCityProduction
discreteEvents.onCityProduction(function(city,item) 
    --civ.ui.text("City production discrete event test")

end)
---&endAutoDoc

---&autoDoc onCityTaken
discreteEvents.onCityTaken(function(city,defender) 
    --civ.ui.text(city.name.." taken from "..defender.name.." discrete event")


end)
---&endAutoDoc

---&autoDoc onScenarioLoaded
discreteEvents.onScenarioLoaded(function() 
    --civ.ui.text("discrete event on scenario loaded")

end)
---&endAutoDoc

---&autoDoc onTurn
discreteEvents.onTurn(function(turn) 
    --civ.ui.text("discrete on turn event 1")
end)
---&endAutoDoc

discreteEvents.onTurn(function(turn) 
    --civ.ui.text("discrete on turn event 2")

end)

discreteEvents.onTurn(function(turn) 
    --civ.ui.text("discrete on turn event 3")
end)

---&autoDoc onUnitKilled
discreteEvents.onUnitKilled(function(loser,winner,aggressor,victim,loserLocation,winnerVetStatus,loserVetStatus) 
    if _global.eventTesting then
        civ.ui.text(loser.type.name.." was killed by "..winner.type.name.." discrete event 1")
    end
end)
---&endAutoDoc

---&autoDoc onUnitDefeated
discreteEvents.onUnitDefeated(function(loser,winner,aggressor,victim,loserLocation,winnerVetStatus,loserVetStatus) 
    if _global.eventTesting then
    civ.ui.text(loser.type.name.." was defeated (possibly by event) by "..winner.type.name.." discrete event")
    end
end)
---&endAutoDoc

discreteEvents.onTurn(function(turn)
    --civ.ui.text("discrete on turn event 4")
end)

---&autoDoc onCityProcessingComplete
discreteEvents.onCityProcessingComplete(function(turn,tribe) 
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onCityProcessingComplete for "..tribe.name.." on turn "..tostring(turn))
    end
end)
---&endAutoDoc

---&autoDoc onTribeTurnBegin
discreteEvents.onTribeTurnBegin(function(turn,tribe) 
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onTribeTurnBegin for "..tribe.name.." on turn "..tostring(turn))
    end
end)
---&endAutoDoc

---&autoDoc onTribeTurnEnd
discreteEvents.onTribeTurnEnd(function(turn,tribe) 
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onTribeTurnEnd for "..tribe.name.." on turn "..tostring(turn))
    end
end)
---&endAutoDoc


discreteEvents.onUnitKilled(function(loser,winner,aggressor,victim,loserLocation,winnerVetStatus,loserVetStatus) 
    if _global.eventTesting then
        civ.ui.text(loser.type.name.." was killed by "..winner.type.name.." discrete event 2")
    end
end)

---&autoDoc onCentauriArrival
discreteEvents.onCentauriArrival(function(tribe)
    --civ.ui.text(tribe.name.." arrived at centauri discrete event")
end)
---&endAutoDoc

---&autoDoc onGameEnds
-- On Game Ends
-- Return true if the game ends as normal,
-- and false otherwise.
-- This is combined with the consolidated events and the
-- legacy events, as well as a separate onGameEnds.lua file
-- If any of these return false, then game end is prevented
-- Not documented or experimented with much
-- based on legacy event engine code, reason is an integer
discreteEvents.onGameEnds(function(reason)
    -- return false to stop the game from ending
    return true
end)
---&endAutoDoc

---&autoDoc onSchism
-- On Schism 
-- Return true (default) if the tribe can schism,
-- and false otherwise.
-- This is combined with the consolidated events and the
-- legacy events, as well as a separate onSchism.lua file
-- If any of these return false, then schism is prevented
discreteEvents.onSchism(function(tribe)

    return true
end)
---&endAutoDoc

---&autoDoc onNegotiation
-- On Negotiation 
-- Return true if the talker can contact the listener,
-- and false otherwise.
-- This is combined with the consolidated events and the
-- legacy events, as well as a separate onNegotiation.lua file
-- If any of these return false, then negotiation is prevented
discreteEvents.onNegotiation(function(talker,listener)

    return true
end)
---&endAutoDoc

---&autoDoc onCanFoundCity
-- Checking if a unit can found a city
-- Return true if the unit can found a city
-- return false if it can't
-- If any one of the consolidated event, the discrete events, 
-- or the separate file event return false, then the city
-- can't be built
-- Notes: Returning true does NOT override any normal city
-- building condition (like no adjacent cities, or cities at sea)
-- Registers a function that is called to determine if `unit` can found 
-- a city at the unit's location. `advancedTribe` is `true` when picking 
-- up a hut with `unit` triggers an advanced tribe. 
-- Return `true` to allow, `false` to disallow.
discreteEvents.onCanFoundCity(function(unit,advancedTribe)
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onCanFoundCity for "..unit.type.name)
    end
    return true
end)
---&endAutoDoc

---&autoDoc onEnterTile
-- onEnterTile(unit,previousTile)
-- executes when a unit successfully enters a tile (so not when it attacks
-- a unit or fails to enter a tile because it lacks movement points)
discreteEvents.onEnterTile(function(unit,previousTile,previousDomainSpec)
    -- if _global.eventTesting then
        -- civ.ui.text("discreteEvents.onEnterTile: "..unit.type.name.." has entered tile ("..text.coordinates(unit.location)..") from tile ("..text.coordinates(previousTile)..").")
    -- end

	
	-- =========================
	-- STACKING PENALTIES (base code by Prof. Garfield) ((see onEnterTile.lua STACKING LIMITATIONS section if you prefer to implement non-damage penalty instead (currently default feature))
	-- ALSO REQUIRES 2 FOLLOWING LINES OF CODE IN HEADER SECTION OF THIS FILE
	-- Inflicts HP damages on unit when human player attempts to enter a tile where it would exceed stacking limit
	-- local unitData = require("unitData")
	-- unitData.defineFlag("stackPenaltyNotApplied",true,"onTribeTurnEnd")
	
	-- Different terrain types have different stacking limits 
	-- DUKW, Engineers, Generals, Supply Hub and Supply Truck units do NOT count against the stacking limit

	-- if flag.value("Stacking Limit Options") == true then 		
	
		-- if unit.owner.isHuman == true and flag.value("Stacking Limit Options") == true then	
			-- if unit.owner == object.pAllies and unit.type.domain == 0 and unitData.flagGetValue(unit,"stackPenaltyNotApplied") then
				-- local totalUnits = 0
				-- local stackingPenaltyMoveCost = 1
				-- for unitOnTile in unit.location.units do
					-- if unitOnTile.type.domain == 0 and not tables.excludeFromTerrainOccupancyLimit[unitOnTile.type.id] 
					-- --and unit.location.baseTerrain ~= object.bTownCenter 
					-- then
						-- totalUnits = totalUnits+1
					-- end
				-- end
				-- local terrainLimit = tables.terrainOccupancyLimit[unit.location.baseTerrain.type]
				-- if totalUnits > terrainLimit then
					-- -- civ.ui.text("The unit on ("..text.coordinates(unit.location)..") is overstacked. The stacking limit for Apennines or Swamp terrain is 3!")
					-- civ.ui.text(func.splitlines("The " .. unit.type.name .. " trying to move into ("..text.coordinates(unit.location)..") caused a massive tracking jam and was forced to retreat to its previous location having expended " .. stackingPenaltyMoveCost .. " extra MP's and suffered " .. param.stackingLimitPenalty .. " HP in the process.\n^\r^The stacking limit for "..unit.location.baseTerrain.name.." terrain is "..tostring(terrainLimit).."!"))
					-- unit:teleport(previousTile)
					-- unit.moveSpent = unit.moveSpent + totpp.movementMultipliers.aggregate*stackingPenaltyMoveCost
					-- gen.setMoved(unit)
					-- unit.damage = math.min(unit.damage + param.stackingLimitPenalty,unit.type.hitpoints-1)
					-- unitData.flagSetFalse(unit,"stackPenaltyNotApplied")	
				-- end	
			-- end	
		-- end	
	-- end
	-- end STACKING PENALTIES =========================

	-- RIVER CROSSING BLOWN WHEN ALLIED UNIT ENTERS RIVER CROSSING TILE THAT HAS BEEN RIGGED TO BLOW (replaces Pioniere event in onUnitKilled.lua)
	-- NEED TO ADD THE LINE BELOW IN HEADER OF THIS FILE
	-- local tileData = require("tileData")
	-- Also check SET RIVER CROSSING TILES TO BE RIGGED TO BLOW BRIDGES in onTurn.lua	
    if tileData.flagGetValue(unit.location,"riggedBridge") and unit.owner == object.pAllies and unit.type.domain == 0 then
		-- Northern River crossing with Crossing Temp 'resource' added to base terrain tile
        if unit.location.terrain == object.tCrossingTemp then
			civ.playSound("LARGEXPL.wav")

			unit.location.baseTerrain = object.bRiverbed
			local tileToChange = civ.getTile(unit.location.x,unit.location.y,unit.location.z)
			tileToChange.terrain = object.tBlownBridge2		
			gen.removeRailroad(unit.location)
			gen.removeRoad(unit.location)
			gen.chartTruthfully(unit.location,unit.owner)
			tileData.flagReset(unit.location,"riggedBridge")
			--tileData.flagSetFalse(unit.location,"riggedBridge")
			unit:teleport(previousTile)
			civ.ui.redrawTile(unit.location)
			text.simple("The bridge on the river crossing located on tile \n^(" .. text.coordinates(unit.location) .. ") has been destroyed by enemy pioniere!\n^\r^It will require you to send an engineer unit to the\n^location if you wish to rebuild the river crossing.","Bridge Destroyed!",object.mBridgeDestroyed)    
		
		-- Southern River crossing with base terrain tile only		
		elseif unit.location.baseTerrain == object.bRiverCrossing and unit.type.domain == 0 then 
			civ.playSound("LARGEXPL.wav")
			
			unit.location.baseTerrain = object.bRiverbed
			local tileToChange = civ.getTile(unit.location.x,unit.location.y,unit.location.z)
			tileToChange.terrain = object.tBlownBridge1		
			gen.removeRailroad(unit.location)
			gen.removeRoad(unit.location)
			gen.chartTruthfully(unit.location,unit.owner)
			tileData.flagReset(unit.location,"riggedBridge")	
			--tileData.flagSetFalse(unit.location,"riggedBridge")
			unit:teleport(previousTile)
			civ.ui.redrawTile(unit.location)
			text.simple("The bridge on the river crossing located on tile \n^(" .. text.coordinates(unit.location) .. ") has been destroyed by enemy pioniere!\n^\r^It will require you to send an engineer unit to the\n^location if you wish to rebuild the river crossing.","Bridge Destroyed!",object.mBridgeDestroyed)    

		end
	end
	
	-- ALLIED UNIT SUFFER DAMAGE AND LOSE REMAINING MP'S WHEN ENTERING PRESET MINEFIELD TILE
	local hpMinPct, hpMaxPct = 0, 0
	local damagePercent = 0
	if tileData.flagGetValue(unit.location,"placedMinenfeld") then
		for unitOnTile in unit.location.units do
			if unit.owner == object.pAllies and unit.type.domain == 0 
			-- Leaders are not affected by minefields
			and (unitOnTile.type ~= object.uGenMontgomery) and (unitOnTile.type ~= object.uGenLeese) and (unitOnTile.type ~= object.uGenMcCreery) and (unitOnTile.type ~= object.uGenPatton) and (unitOnTile.type ~= object.uGenClark) and (unitOnTile.type ~= object.uGenTruscott)
			-- and (unitOnTile.type ~= object.uUSTruck) and (unitOnTile.type ~= object.uUKTruck) -- Prevent the minenfield events for trucks due to the complication related to units they transport
			then
				civ.playSound("LARGEXPL.wav")
				tileData.flagReset(unit.location,"placedMinenfeld")
				-- unit.damage = math.min(unit.damage + param.minenfeldPenalty,unit.type.hitpoints-1)
				-- Determine the min and max damage %
				hpMinPct, hpMaxPct = 20, 50	
				-- 
				if hpMaxPct > 0 then
					local damagePercent = math.random(hpMinPct, hpMaxPct)
					local damageAmount = round(unit.type.hitpoints * (damagePercent / 100)) -- Also see "local function round (decimal) above"
					local oldHP = unit.hitpoints
					if damageAmount >= oldHP and (unit.type ~= object.uUSTruck) and (unit.type ~= object.uUKTruck) then -- Added truck condition here to prevent the minenfield events from killing trucks due to the complication related to units they transport
						-- print("Unit damage: " .. unit.type.name .. " at " .. unit.x .. "," .. unit.y .. "," .. unit.z .. " KILLED due to HP being reduced from " .. oldHP .. " HP to 0 HP!")
						text.simple("Your " .. unit.type.name .. " unit entered a minefield on tile (" .. text.coordinates(unit.location) .. ") and was KILLED due to HP being reduced from " .. oldHP .. " HP to 0 HP!","Minefield Clearance",object.mMinefield)    
						civ.deleteUnit(unit)
						return true -- This line is required to prevent 'tile coordinate' error (works with 'local function onEnterTile(unit,previousTile,previousDomainSpec)' correction made by Prof. Garfield in events.lua)
					elseif damageAmount > 0 then
						local newHP = oldHP - damageAmount
						-- print("Unit damage: " .. unit.type.name .. " at " .. unit.x .. "," .. unit.y .. "," .. unit.z .. " reduced from " .. oldHP .. "HP to " .. newHP .. "HP")
						text.simple("Your " .. unit.type.name .. " unit entered a minefield on tile (" .. text.coordinates(unit.location) .. ") and has suffered casualties as a result! Sappers attached to the unit are dispatched to clear the area of mines and therefore the unit ends its movement!","Minefield Clearance",object.mMinefield)    
						unit.damage = unit.damage + damageAmount
						unit.moveSpent = 255
					end
				end	

			end
		end
	end		
	
	
end)
---&endAutoDoc

---&autoDoc onFinalOrderGiven
-- onFinalOrderGiven(unit)
-- executes when a unit has been given its final order for the turn.
-- that is, when a new unit is active and the previous unit has spent
-- all its movement points
discreteEvents.onFinalOrderGiven(function(unit)
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onFinalOrderGiven: "..unit.type.name.." has been given its order.")
    end
end)
---&endAutoDoc

-- Key press events are probably best registered in
-- MechanicsFiles\keyPressSettings.lua
-- This is here to facilitate documentation
---&autoDoc onKeyPress
-- The keyCode is an integer that corresponds to a particular key on the keyboard.
-- The keyboard module provides names for these codes.
discreteEvents.onKeyPress(function(keyCode)
    if _global.eventTesting and keyboard.backspace == keyCode then
        civ.ui.text("discreteEvents.onKeyPress: The backspace key has been pressed.")
    end
end)
---&endAutoDoc

---&autoDoc onCityWindowOpened
-- Executes when a city window is opened.
-- Note that the AI doesn't open city windows.
discreteEvents.onCityWindowOpened(function(city)
    if _global.eventTesting then
        civ.ui.text("discreteEvents.onCityWindowOpened: The city window for "..city.name.." has been opened.")
    end
end)
---&endAutoDoc

-- ===============================================================================
--
--          End of File
--
-- ===============================================================================
--      In order to register discrete events, you don't need 
--      to return anything, but the file must be 'required'
--      by another file.  Discrete Events can be registered in any file,
--      provided it has the following require line:
--      
--      local discreteEvents = require("discreteEventsRegistrar")

local versionTable = {}
gen.versionFunctions(versionTable,versionNumber,fileModified,"MechanicsFiles".."\\".."discreteEvents.lua")
return versionTable

